#include "dbase.h"

static DB_ENV * dbenv = NULL;
static DB * dbp = NULL;

int db_open()
{
    int retval;

    if ((retval = db_env_create(&dbenv, 0)) != 0) {
	fprintf(stderr, "Error creating env handle: %s\n",
		db_strerror(retval));
	return -1;
    }
    
    retval = dbenv->open(dbenv,
			 DB_ENV_HOME,
			 DB_CREATE|DB_PRIVATE|DB_INIT_MPOOL|DB_INIT_LOG,
			 0664);
    if (retval != 0) {
	fprintf(stderr, "Error opening env: %s\n",
		db_strerror(retval));
	return -1;
    }

    if ((retval = db_create(&dbp, dbenv, 0)) != 0) {
	fprintf(stderr, "Error creating db handle: %s\n",
		db_strerror(retval));
	return -1;
    }
  
    dbp->set_errfile(dbp, stderr);
    dbp->set_errpfx(dbp, DB_NAME);
  
    retval = dbp->open(dbp,
		       NULL,
		       DB_NAME,
		       NULL,
		       DB_HASH,
		       DB_CREATE,
		       0664);
   
    if (retval != 0) {
	fprintf(stderr, "Error opening db: %s\n",
		db_strerror(retval));
	return -1;
    }

    return 1;
}

void db_close()
{
    if (dbp)
	/* we sync on every put so no need for sync here 
	 * FIXME remove if using transactions (which we should) */
	dbp->close(dbp, DB_NOSYNC);
    if (dbenv) {
	dbenv->close(dbenv, 0);
    }
}

int db_dev_exists(char * addr)
{
    DBT key, data;
    char entry[18];
    int retval;

    memset(&key, 0, DBT_SIZE);
    memset(&data, 0, DBT_SIZE);
    memset(entry, 0, DBENTRY_SIZE);

    key.data = addr;
    key.size = 18;
    
    data.data = entry;
    data.ulen = DBENTRY_SIZE;
    data.flags = DB_DBT_USERMEM;

    retval = dbp->get(dbp, NULL, &key, &data, 0);
    if (retval == DB_NOTFOUND) 
	return 0;

    printf("found address: %s\n", entry);
    
    return 1;
}

int db_dev_put(char * addr)
{
    DBT key, data;
    char entry[18];
    int retval;

    memset(&key, 0, DBT_SIZE);
    memset(&data, 0, DBT_SIZE);
    memset(entry, 0, DBENTRY_SIZE);

    memcpy(entry, addr, 18);
    key.data = entry;
    key.size = 18;
    
    data.data = entry;
    data.size = DBENTRY_SIZE;

    retval = dbp->put(dbp, NULL, &key, &data, 0);
    if (retval != 0) {
	fprintf(stderr, "Error add %s to database: %s\n",
		addr, db_strerror(retval));
	return 0;
    }
    /* FIXME use transactions instead of syncs here 
     * as sync is much more expensive. */
    dbp->sync(dbp, 0);

    return 1;
}

int db_dev_remove(char * addr)
{
    DBT key;
    int retval;

    memset(&key, 0, DBT_SIZE);
    key.data = addr;
    key.size = 18;
    
    retval = dbp->del(dbp, NULL, &key, 0);
    if (retval != 0) {
	fprintf(stderr, "Error removing %s from database: %s\n",
		addr, db_strerror(retval));
	return 0;
    }
    return 1;
}
