/*
CPRY: (C)2003 Beier & Dauskardt IT (www.bdit.de)
FILE: main.c
DATE: 05.03.2003
AUTH: S.Dauskardt
CMNT: opd (OPEX Push Daemon) - main entry.


TODO:	check if sdp entry is already present.


 - Heavily based on BlueZ + OpenOBEX
 	Thanks to all contributors !
*/
/*
 *     This program is free software; you can redistribute it and/or
 *     modify it under the terms of the GNU General Public License as
 *     published by the Free Software Foundation; either version 2 of
 *     the License, or (at your option) any later version.
 *
 *     This program is distributed in the hope that it will be useful,
 *     but WITHOUT ANY WARRANTY; without even the implied warranty of
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 *     GNU General Public License for more details.
 *
 *     You should have received a copy of the GNU General Public License
 *     along with this program; if not, write to the Free Software
 *     Foundation, Inc., 59 Temple Place, Suite 330, Boston,
 *     MA 02111-1307 USA
 *
 *
 */
const char *VERSION = "0.2";

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <signal.h>
#include <fcntl.h>
#include <syslog.h>
#include <errno.h>
#include <time.h>
#include <fcntl.h>

#include <sys/ioctl.h>
#include <sys/socket.h>
#include <sys/time.h>
#include <sys/stat.h>

#include <bluetooth/sdp.h>
#include <bluetooth/sdp_lib.h>


#include <openobex/obex.h>

#include "lib.h"
#include "obex_handler.h"
#include "dbase.h"

#include "main.h"

// Ugly globals..
//
int		srv_RFCommChannel;
int		srv_Modes;
int		srv_Daemonize;
int		srv_Debug;
int		srv_AddSDP;
int		srv_Metadata;
char		srv_BasePath[256];



// Bits possibly missing in sdp.h
//
#ifndef BIP_SVCLASS_ID
#define BIP_SVCLASS_ID                  0x111b          // Responder
#endif

#ifndef BIP_PROFILE_ID
#define BIP_PROFILE_ID                  0x111a
#endif




int OPD_SocketType()
{
	if( srv_Modes & (OPDM_OBEX|OPDM_BIP) )
		return AF_BLUETOOTH;

	if( srv_Modes & (OPDM_IRDA) )
		return AF_IRDA;

	return 0;
}


//
//
//
sdp_record_t obex_record;

int add_opush(sdp_session_t *session)
{
	sdp_list_t *svclass_id, *pfseq, *apseq, *root;
	uuid_t root_uuid, opush_uuid, l2cap_uuid, rfcomm_uuid, obex_uuid;
	sdp_profile_desc_t profile[1];
	sdp_list_t *aproto, *proto[3];
	uint8_t u8 = srv_RFCommChannel;
	sdp_data_t *channel;

	memset((void *)&obex_record, 0, sizeof(sdp_record_t));
	obex_record.handle = 0xffffffff;
	sdp_uuid16_create(&root_uuid, PUBLIC_BROWSE_GROUP);
	root = sdp_list_append(0, &root_uuid);
	sdp_set_browse_groups(&obex_record, root);

	sdp_uuid16_create(&opush_uuid, OBEX_OBJPUSH_SVCLASS_ID);
	svclass_id = sdp_list_append(0, &opush_uuid);
	sdp_set_service_classes(&obex_record, svclass_id);

	sdp_uuid16_create(&profile[0].uuid, OBEX_OBJPUSH_PROFILE_ID);
	profile[0].version = 0x0100;
	pfseq = sdp_list_append(0, profile);
	sdp_set_profile_descs(&obex_record, pfseq);

	sdp_uuid16_create(&l2cap_uuid, L2CAP_UUID);
	proto[0] = sdp_list_append(0, &l2cap_uuid);
	apseq = sdp_list_append(0, proto[0]);

	sdp_uuid16_create(&rfcomm_uuid, RFCOMM_UUID);
	proto[1] = sdp_list_append(0, &rfcomm_uuid);
	channel = sdp_data_alloc(SDP_UINT8, &u8);
	proto[1] = sdp_list_append(proto[1], channel);
	apseq = sdp_list_append(apseq, proto[1]);

	sdp_uuid16_create(&obex_uuid, OBEX_UUID);
	proto[2] = sdp_list_append(0, &obex_uuid);
	apseq = sdp_list_append(apseq, proto[2]);

	aproto = sdp_list_append(0, apseq);
	sdp_set_access_protos(&obex_record, aproto);
	sdp_data_free(channel);
	sdp_list_free(proto[0], 0);
	sdp_list_free(proto[1], 0);
	sdp_list_free(proto[2], 0);
	sdp_list_free(apseq, 0);
	sdp_list_free(aproto, 0);

	sdp_set_info_attr(&obex_record, "OBEX Object Push", 0, 0);




	/* Add Supported Formats List Attribute (0x0303)
	*/
	{
		uint8_t uint8 = SDP_UINT8;
		uint8_t formats[]={0x01, 0x02, 0x03, 0x04, 0x05, 0x06, 0xFF};
		void **dtds   = (void **)malloc (sizeof (formats) * sizeof(void *));
		void **values = (void **)malloc (sizeof (formats) * sizeof(void *));
		int i;
		sdp_data_t *seq;

		for(i = 0; i < sizeof (formats); i++){
			dtds[i] = &uint8;
			values[i] = &formats [i];
		}
		seq = sdp_seq_alloc (dtds, values, sizeof (formats));
		sdp_attr_add(&obex_record, 0x303, seq);


		free (dtds);
		free (values);
	}




	if (0 > sdp_record_register(session, &obex_record, SDP_RECORD_PERSIST)) {
		syslog(LOG_ERR,"OBEX Object Push Service Record registration failed with sdpd.\n");
		return -1;
	}
	syslog(LOG_INFO,"OBEX Object Push service registered with sdpd\n");
	return 0;
}





//
//
//
sdp_record_t bip_record;

int add_bip(sdp_session_t *session)
{
	sdp_list_t *svclass_id, *pfseq, *apseq, *root;
	uuid_t root_uuid, opush_uuid, l2cap_uuid, rfcomm_uuid, obex_uuid;
	sdp_profile_desc_t profile[1];
	sdp_list_t *aproto, *proto[3];
	uint8_t u8 = srv_RFCommChannel;
	sdp_data_t *channel;

	memset((void *)&bip_record, 0, sizeof(sdp_record_t));
	bip_record.handle = 0xffffffff;
	sdp_uuid16_create(&root_uuid, PUBLIC_BROWSE_GROUP);
	root = sdp_list_append(0, &root_uuid);
	sdp_set_browse_groups(&bip_record, root);

	sdp_uuid16_create(&opush_uuid, BIP_SVCLASS_ID);
	svclass_id = sdp_list_append(0, &opush_uuid);
	sdp_set_service_classes(&bip_record, svclass_id);

	sdp_uuid16_create(&profile[0].uuid, BIP_PROFILE_ID);
	profile[0].version = 0x0100;
	pfseq = sdp_list_append(0, profile);
	sdp_set_profile_descs(&bip_record, pfseq);

	sdp_uuid16_create(&l2cap_uuid, L2CAP_UUID);
	proto[0] = sdp_list_append(0, &l2cap_uuid);
	apseq = sdp_list_append(0, proto[0]);

	sdp_uuid16_create(&rfcomm_uuid, RFCOMM_UUID);
	proto[1] = sdp_list_append(0, &rfcomm_uuid);
	channel = sdp_data_alloc(SDP_UINT8, &u8);
	proto[1] = sdp_list_append(proto[1], channel);
	apseq = sdp_list_append(apseq, proto[1]);

	sdp_uuid16_create(&obex_uuid, OBEX_UUID);	// nicht BIP !!!
	proto[2] = sdp_list_append(0, &obex_uuid);
	apseq = sdp_list_append(apseq, proto[2]);

	aproto = sdp_list_append(0, apseq);
	sdp_set_access_protos(&bip_record, aproto);
	sdp_data_free(channel);
	sdp_list_free(proto[0], 0);
	sdp_list_free(proto[1], 0);
	sdp_list_free(proto[2], 0);
	sdp_list_free(apseq, 0);
	sdp_list_free(aproto, 0);

	sdp_set_info_attr(&bip_record, "BIP", 0, 0);



	/* Add Supported capabilities (0x0310) 
	*/
	{
		uint8_t uint8		= SDP_UINT8;
	        uint8_t capabilities	= 0x01;
		void *dtds[1];
		void *values[1];
		sdp_data_t *seq;

		dtds[0]   = &uint8;
		values[0] = &capabilities;

		seq = sdp_seq_alloc (dtds, values, 1);		//sizeof (formats));
		sdp_attr_add(&bip_record, 0x310, seq);
	}
	{
		uint16_t uint16		= SDP_UINT16;
	        uint16_t capabilities	= 0x01;
		void *dtds[1];
		void *values[1];
		sdp_data_t *seq;

		dtds[0]   = &uint16;
		values[0] = &capabilities;

		seq = sdp_seq_alloc (dtds, values, 1);
		sdp_attr_add(&bip_record, 0x311, seq);
	}
	{
		uint32_t uint32		= SDP_UINT32;
	        uint32_t capabilities	= 0x03;
		void *dtds[1];
		void *values[1];
		sdp_data_t *seq;

		dtds[0]   = &uint32;
		values[0] = &capabilities;

		seq = sdp_seq_alloc (dtds, values, 1);
		sdp_attr_add(&bip_record, 0x312, seq);
	}
	{
		uint64_t uint64		= SDP_UINT64;
	        uint64_t capabilities	= 8*1024*1024;
		void *dtds[1];
		void *values[1];
		sdp_data_t *seq;

		dtds[0]   = &uint64;
		values[0] = &capabilities;

		seq = sdp_seq_alloc (dtds, values, 1);
		sdp_attr_add(&bip_record, 0x313, seq);
	}





	if (0 > sdp_record_register(session, &bip_record, SDP_RECORD_PERSIST)) {
		syslog(LOG_ERR,"BIP Service Record registration failed with sdpd.\n");
		return -1;
	}
	syslog(LOG_INFO,"BIP service registered with sdpd\n");
	return 0;
}


/* Add local service
*/
int AddSDPService()
{
	int ret = 0;

	sdp_session_t *sess = sdp_connect(BDADDR_ANY, BDADDR_LOCAL, 0);
	if (!sess){
		syslog(LOG_ERR, "Failed to connect to SDP server.\n");
		return -1;
	}

	if( srv_Modes & OPDM_OBEX ){
		ret = add_opush(sess);
	}

	if( srv_Modes & OPDM_BIP ){
		ret = add_bip(sess);
	}

	sdp_close(sess);
	return ret;
}

int RemoveSDPService()
{
	int	ret=0;

	sdp_session_t *sess = sdp_connect(BDADDR_ANY, BDADDR_LOCAL, 0);
	if (!sess){
		syslog(LOG_ERR, "Failed to connect to SDP server.\n");
		return -1;
	}


	if( srv_Modes & OPDM_OBEX ){
		ret = sdp_record_unregister(sess, &obex_record);
	}

	if( srv_Modes & OPDM_BIP ){
		ret = sdp_record_unregister(sess, &bip_record);
	}

	if( 0 > ret ){
		syslog(LOG_ERR,"Service Record unregistration failed with sdpd.\n");
	}else{
		syslog(LOG_INFO,"service unregistered with sdpd\n");
	}

	sdp_close(sess);
	return ret;
}




void sig_term(int sig)
{
	syslog(LOG_INFO,"Terminating .. \n");

	if( srv_AddSDP )
		RemoveSDPService();

	db_close();
	// TODO: OBEX_Cleanup ??

	exit(0);
}

// void obex_event(obex_t *handle, obex_object_t *object, int mode, int event, int obex_cmd, int obex_rsp);


void printusage()
{
	printf("Basic Imaging Profile - Image responder daemon (c)2003 bdit.de\n");
    	printf("Usage:\n");
	printf("opd [option [arg]] [option [arg]]...\n");
	printf("\n");
	printf("    options:\n");
	printf("\n");
	printf("	--mode < BIP | OBEX | BIP+OBEX | IRDA >\n");
	printf("	Run in plain OBEX, extended BIP mode, both or IrDA (default: OBEX).\n");
	printf("\n");
	printf("	--path <target-path>\n");
	printf("	Path where to save recieved files to (default: /tmp)\n");
	printf("\n");
	printf("	--channel <channel>\n");
	printf("	RFCOMM channel.\n");
	printf("\n");
	printf("	--sdp\n");
	printf("	Register with sdpd.\n");
	printf("\n");
	printf("	--metadata\n");
	printf("	Save metadata with recieved files.\n");
	printf("\n");
	printf("	--daemonize\n");
	printf("	Run in background.\n");
	printf("\n");
}

int main(int argc, char **argv,char *env[])
{
	obex_t		*server_handle = NULL;
	int		err;
	int		fd,i;

	if( argc < 2 ){
		printusage();
		exit(0);
	}

	srv_RFCommChannel	= 1;
	srv_Modes		= OPDM_OBEX;
	srv_Daemonize		= 0;
	srv_Debug		= 0;
	srv_AddSDP		= 0;
	srv_Metadata		= 0;
	strcpy(srv_BasePath,"/tmp/");


	// Read commandline.
	//
	for(i=1;i<argc;i++){

		if(!strcasecmp(argv[i],"--mode")){
			if(i+1<argc){
				if(!strcasecmp(argv[i+1],"BIP")){
					srv_Modes = OPDM_BIP;
					i++;
					continue;
				} else if(!strcasecmp(argv[i+1],"OBEX")){
					srv_Modes = OPDM_OBEX;
					i++;
					continue;
				} else if(!strcasecmp(argv[i+1],"OBEX+BIP")){
					srv_Modes = OPDM_OBEX|OPDM_BIP;
					i++;
					continue;
				} else if(!strcasecmp(argv[i+1],"IRDA")){
					srv_Modes = OPDM_IRDA;
					i++;
					continue;
				}
			}
		}

		if(!strcasecmp(argv[i],"--path")){
			if(i+1<argc){
				int len;

				strcpy(srv_BasePath,argv[i+1]);
				len = strlen(srv_BasePath);
				if(len){
					if( srv_BasePath[len-1] != '/' )
						strcat(srv_BasePath,"/");
				}else{
					strcpy(srv_BasePath,"./");
				}

				i++;
				continue;
			}
		}

		if(!strcasecmp(argv[i],"--channel")){
			if(i+1<argc){
				srv_RFCommChannel = atoi(argv[i+1]);
				i++;
				continue;
			}
		}

		if(!strcasecmp(argv[i],"--daemonize")){
			srv_Daemonize = 1;
			continue;
		}

		if(!strcasecmp(argv[i],"--sdp")){
			srv_AddSDP = 1;
			continue;
		}

		if(!strcasecmp(argv[i],"--metadata")){
			srv_Metadata = 1;
			continue;
		}

		if(!strcasecmp(argv[i],"--debug")){
			srv_Debug = 1;
			continue;
		}





		printusage();
		exit(0);
	}


	if( srv_Metadata && (srv_Modes & OPDM_IRDA) ){
		printf("--metadata is not supported with IrDA\n");
		exit(0);
	}



	// run in background ?
	//
	if( srv_Daemonize ){
		if( fork() )
			exit(0);

		/* Direct stdin,stdout,stderr to '/dev/null'
		*/
		if( srv_Debug ){

			fd = open("/dev/null", O_RDWR);
			dup2(fd, 0);
			close(fd);

			fd = open("_opd_msgs.txt", O_RDWR | O_CREAT);
			dup2(fd, 1);
			close(fd);

			fd = open("_opd_errs.txt", O_RDWR | O_CREAT);
			dup2(fd, 2);
			close(fd);

		}else{

			fd = open("/dev/null", O_RDWR);
			dup2(fd, 0); dup2(fd, 1); dup2(fd, 2);
			close(fd);
		}

		setsid();

		chdir("/");
	}

	umask(0);		// 0077 ??

	init_title(argc, argv, env, "opd: ");
	set_title("initializing");

	/* Start logging to syslog and stderr
	*/
	openlog("opd", LOG_PID | LOG_NDELAY | LOG_PERROR, LOG_DAEMON);
	syslog(LOG_INFO, "OBEX Push Daemon - version %s started", VERSION);




	// Anything Bluetooth ?
	//
	switch( OPD_SocketType() ){
	case AF_BLUETOOTH:

		// Register sdp entry ??
		//
		if( srv_AddSDP )
			AddSDPService();


		server_handle = OBEX_Init(OBEX_TRANS_BLUETOOTH, Handle_OBEX_Event, OBEX_FL_KEEPSERVER);
		if(!server_handle){
			syslog(LOG_ERR, "Failed to initialize Bluetooth OBEX.\n");
			return 1;
		}


		err = BtOBEX_ServerRegister(server_handle, NULL, srv_RFCommChannel);
		if(err < 0){
			syslog(LOG_ERR,"Failed to register BtOBEX server on RFCOMM channel %d. (%d / %d)\n",
				srv_RFCommChannel,
				err,
				errno
			);
			return 1;
		}

		switch( srv_Modes & (OPDM_OBEX|OPDM_BIP) ){
		case OPDM_OBEX:
			syslog(LOG_ERR, "OBEX server listening on channel %d\n",srv_RFCommChannel);
			set_title("waiting for incomming OBEX connections on channel %d...",srv_RFCommChannel);
			break;
		case OPDM_BIP:
			syslog(LOG_ERR, "BIP server listening on channel %d\n",srv_RFCommChannel);
			set_title("waiting for incomming BIP connections on channel %d...",srv_RFCommChannel);
			break;
		case OPDM_OBEX|OPDM_BIP:
			syslog(LOG_ERR, "OBEX+BIP server listening on channel %d\n",srv_RFCommChannel);
			set_title("waiting for incomming OBEX+BIP connections on channel %d...",srv_RFCommChannel);
			break;
		}
		break;
		
	case AF_IRDA:
		printf("Using IrDA transport\n");

		server_handle = OBEX_Init(OBEX_TRANS_IRDA, Handle_OBEX_Event, OBEX_FL_KEEPSERVER);
		if(!server_handle){
			syslog(LOG_ERR, "Failed to initialize IrDA OBEX.\n");
			return 1;
		}

		err = IrOBEX_ServerRegister(server_handle, "OBEX");
		if(err < 0){
			syslog(LOG_ERR,"Failed to register IrDA OBEX server  (%d / %d)\n",
				err,
				errno
			);
			return 1;
		}

		syslog(LOG_ERR, "OBEX server listening via IrDA.\n");
		set_title("waiting for incomming OBEX connections via IrDA...");
		break;

	default:
		syslog(LOG_ERR, "Can't handle internal mode 0x%02x\n",srv_Modes);
		return 0;
	}


	//
	//
	signal(SIGINT,  sig_term);
	signal(SIGTERM, sig_term);
	signal(SIGABRT, sig_term);
	signal(SIGQUIT, sig_term);

	db_open();
	db_dev_put("00:0F:DE:0E:E2:A8");
	//db_dev_put("00:02:72:B2:A3:E5");	
	// Action.
	//
	for(;;){
		// Minute...
		OBEX_HandleInput(server_handle, 60);

		if( srv_Debug ){
			printf(" -- MARK -- \n");
			fprintf(stderr," -- MARK -- \n");
			fflush(stdout);
			fflush(stderr);
		}
	}
	
	printf("opd: left loop ??\n");
	db_close(); /* shouldn't get here, but you never know */
}
