/*
CPRY: (C)2003 Beier & Dauskardt IT (www.bdit.de)
FILE: obex_handler.c
AUTH: Sancho Dauskardt
DATE: 05.03.2003
CMNT: opd OBEX Event processing.
*/

/*
 *
 *
 *	Parts by:       Pontus Fuchs <pontus.fuchs@tactel.se>
 *
 *
 *     This program is free software; you can redistribute it and/or
 *     modify it under the terms of the GNU General Public License as
 *     published by the Free Software Foundation; either version 2 of
 *     the License, or (at your option) any later version.
 *
 *     This program is distributed in the hope that it will be useful,
 *     but WITHOUT ANY WARRANTY; without even the implied warranty of
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 *     GNU General Public License for more details.
 *
 *     You should have received a copy of the GNU General Public License
 *     along with this program; if not, write to the Free Software
 *     Foundation, Inc., 59 Temple Place, Suite 330, Boston,
 *     MA 02111-1307 USA
 *
 *
 */
#include <stdio.h>
#include <stdlib.h>
#include <sys/stat.h>
#include <sys/file.h>
#include <unistd.h>
#include <fcntl.h>
#include <string.h>
#include <syslog.h>

#include <sys/socket.h>
#include <bluetooth/bluetooth.h>
#include <bluetooth/rfcomm.h>
#include <openobex/obex.h>

#include <asm/types.h>
#include <linux/socket.h>
#include <linux/irda.h>


#include "obex_handler.h"
#include "dbase.h"
#include "main.h"

#define SPAM_LOCK "/tmp/spam.lock"
#define OPD_LOCK "/tmp/opd.lock"


#define TRUE  1
#define FALSE 0

static int isLocked() {
    struct stat st;
    int s_lock = stat(SPAM_LOCK, &st);
    if (s_lock > 0)
	return 1;
    return  0;
}

static int acquireLock(int timeout) {
    struct stat st;
    struct timeval stv, ctv;
    int locked;
    int fd;

    gettimeofday(&stv, NULL);

    do {
	int s_lock = stat(SPAM_LOCK, &st);
	if (s_lock >= 0)
	    locked = 1;
	else 
	    locked = 0;
	gettimeofday(&ctv, NULL);
	if ((ctv.tv_sec - stv.tv_sec) >= timeout) 
	    return 0;
	usleep(5000);
    } while(locked);
    
    fd = open(OPD_LOCK, O_RDONLY|O_CREAT);
    close(fd);
    return 1;
}

static void unLock() {
    unlink(OPD_LOCK); 
}

int	srv_BIPNextImageHandle=1;



// Answer for BIP GetCapabilities(): 'x-bt/img-capabilities'
//
const char *bt_img_capabilities =
	"<imaging-capabilities version=\"1.0\">\n"
	"	<image-formats encoding=\"JPEG\" pixel=\"640*480-4000*3000\"/>\n"
	"</imaging-capabilities>\n"
;





/* Get the filesize in a "portable" way
*/
int get_filesize(const char *filename)
{
	struct stat stats;
	/*  Need to know the file length */
	stat(filename, &stats);
	return (int) stats.st_size;
	return 0;
}


/*
** Read a file and alloc a buffer for it
*/
uint8_t* easy_readfile(const char *filename, int *file_size)
{
	int actual;
	int fd;
	uint8_t *buf;

	*file_size = get_filesize(filename);
	printf("name=%s, size=%d\n", filename, *file_size);

	fd = open(filename, O_RDONLY, 0);
	if (fd == -1) {
		return NULL;
	}
	
	buf = (uint8_t *)malloc(*file_size);
	if(!buf){
		return NULL;
	}

	actual = read(fd, buf, *file_size);
	close(fd);

	*file_size = actual;
	return buf;
}



/*
 *
 *
 *    First remove path and add "/tmp/". Then save.
 *
 */
int save_binary_file(char *filename, char *actname, const uint8_t *buf, int len)
{
	int fd;
	int actual;

	printf("Filename = %s\n", actname);

	if (!strcasecmp(actname, "add_admin.cmd")) {
	    unsigned char * ptr = (unsigned char *)buf;
	    int i;
	    printf("len %d\n", len);
	    if (db_dev_exists((unsigned char *)buf))
		return 0;
	    printf("add: %s\n", db_dev_put((unsigned char *)buf)?"succeeded":"failed");
	    actual = len; 
	} else if (!strcasecmp(actname, "rem_admin.cmd")) {
	    unsigned char * ptr = (unsigned char *)buf;
	    int i;
	    printf("len %d\n", len);
	    if (!db_dev_exists((unsigned char *)buf))
		return 0;
	    printf("del: %s\n", db_dev_remove((unsigned char *)buf)?"succeeded":"failed");
	    actual = len; 
	} else {	
	    fd = open(filename, O_RDWR | O_CREAT, DEFFILEMODE);
	    if ( fd < 0) {
		    perror( filename);
		    return -1;
	    }

	    actual = write(fd, buf, len);
	    close(fd);

	    printf( "Wrote %s (%d bytes)\n", filename, actual);
	}
	return actual;
}










//
//
void dump_hex(void *dataptr,int length)
{
	int	i;

	char	ascii[17];
	int	asciiofs=0;

	unsigned char	*data = (unsigned char *)dataptr;

	for(i=0;i<length;i++){
		if(i % 16 == 0)
			printf("%02x:",i);

		printf(" %02x",data[i]);

		if( data[i] >= 0x20 ){
		    ascii[asciiofs++] = data[i];
		}else{
		    ascii[asciiofs++] = '.';
		}

		if(i % 16 == 15){
			ascii[asciiofs] =0;
			printf("   %s\n",ascii);
			asciiofs=0;
		}
	}

	if(i % 16 != 0){
		for(;(i%16!=0);i++){
    			printf("   ");
			ascii[asciiofs++] = ' ';
		}

		ascii[asciiofs] =0;
		printf("   %s\n",ascii);
	}
}











// Get the Bluetooth bdaddr to an OBEX-handle.
//
int Get_handle_bdaddr(obex_t *handle,char *str)
{
	int			fd;
	struct	sockaddr_rc	sa;
	socklen_t		len = sizeof(sa);

	fd = OBEX_GetFD(handle);
	if( fd < 0 )
		return 0;

	if( getpeername(fd,(struct sockaddr *)&sa,&len) < 0 )
		return 0;

	if( sa.rc_family != AF_BLUETOOTH )
		return 0;

	ba2str(&sa.rc_bdaddr,str);
	return 1;
}

int Get_handle_irdaaddr(obex_t *handle,char *str)
{
	int			fd;
	struct	sockaddr_irda	sa;
	socklen_t		len = sizeof(sa);

	fd = OBEX_GetFD(handle);
	if( fd < 0 )
		return 0;

	if( getpeername(fd,(struct sockaddr *)&sa,&len) < 0 )
		return 0;

	if( sa.sir_family != AF_IRDA )
		return 0;

	sprintf(str,"%02x-%08x-%s",
		sa.sir_lsap_sel,	/* LSAP selector */
		sa.sir_addr,		/* Device address */
		sa.sir_name		/* Usually <service>:IrDA:TinyTP */
	);

	return 1;
}


// For log-messages.
//
int GetOBEXPeerName(obex_t *handle,char *str)
{
	switch( OPD_SocketType() ){
	case AF_BLUETOOTH:
		if( Get_handle_bdaddr(handle,str) )
			return 1;

		strcpy(str,"unknown-br");
		break;

	case AF_IRDA:
		if( Get_handle_irdaaddr(handle,str) )
			return 1;

		strcpy(str,"unknown-irda");
		break;

	default:
		strcpy(str,"unknown-af");
		break;
	}
	return 0;
}











/* void Handle_OBEX_CMD_PUT(obex_t *handle, obex_object_t *object)
 *
 *
 *
 *    Parse what we got from a PUT
 *
 */
void Handle_OBEX_CMD_PUT(obex_t *handle, obex_object_t *object)
{
	obex_headerdata_t	hv;
	uint8_t			hi;
	int			hlen;

	const uint8_t		*body = NULL;
	int			body_len = 0;

	const uint8_t		*headertype = NULL;
	int			headertype_len = 0;

	const uint8_t		*imgdesc = NULL;
	int			imgdesc_len = 0;

	int			GotName=0;
	char			NameBuf[1024];

	char			*LastSlash;
	char			SafeName[1024];

	int			fd;



	while(OBEX_ObjectGetNextHeader(handle, object, &hi, &hv, &hlen))	{
		switch(hi)	{
		case OBEX_HDR_BODY:
			body = hv.bs;
			body_len = hlen;
			break;

		case OBEX_HDR_NAME:
			if( hlen/2 < sizeof(NameBuf) ){
				OBEX_UnicodeToChar(NameBuf, hv.bs, hlen);
				GotName = TRUE;
			}
			break;

		case OBEX_HDR_LENGTH:
			printf("HEADER_LENGTH = %d\n", hv.bq4);
			break;


		case OBEX_HDR_TYPE:
			printf("HEADER_TYPE =  '%s' #%d  \n", hv.bs, hlen);

			dump_hex((void *)hv.bs,hlen);

			headertype = hv.bs;
			headertype_len = hlen;
			break;

		case 0xcb:
			printf("CONNECTION_ID = %#x (%d length)\n", hv.bq4,hlen);
			break;

		case 0x71:
			printf("IMAGE_DESCRIPTION = ? (%d length)\n",hlen);

			dump_hex((void *)hv.bs,hlen);

			imgdesc = hv.bs;
			imgdesc_len = hlen;
			break;


		default:
			printf("%s() Skipped header %02x\n", __FUNCTION__, hi);
		}
	}

	if(!body){
		printf("Got a PUT without a body\n");
		OBEX_ObjectSetRsp(object, OBEX_RSP_CONTINUE, OBEX_RSP_SUCCESS);
		return;
	}

	if(!GotName){
		sprintf(NameBuf,"OBEX_PUT_Unknown_object_%07d",srv_BIPNextImageHandle);
		GotName = TRUE;
		printf("Got a PUT without a name. Setting name to %s\n", NameBuf);
	}


	// Remove possible path from target name und
	// use our path.
	//
	strcpy(SafeName,srv_BasePath);

	if( (LastSlash = strrchr(NameBuf,'/')) ){
		LastSlash++;
	}else{
		if( (LastSlash = strrchr(NameBuf,'\\')) ){
			LastSlash++;
		}else{
			LastSlash = NameBuf;
		}
	}

	strcat(SafeName,LastSlash);




	if( srv_Metadata && (OPD_SocketType()==AF_BLUETOOTH) ){

		char	bdaddr_str[256];

		if( headertype ){
			char	DumpName[2048];
			sprintf(DumpName,"%s.bt-headertype",SafeName);
			save_binary_file(DumpName, LastSlash, headertype, headertype_len);
		}

		if( imgdesc ){
			char	DumpName[2048];
			sprintf(DumpName,"%s.bt-imagedesc",SafeName);
			save_binary_file(DumpName, LastSlash, imgdesc, imgdesc_len);
		}


		if( Get_handle_bdaddr(handle,bdaddr_str)){
			char	DumpName[2048];
			sprintf(DumpName,"%s.bt-addr",SafeName);
			save_binary_file(DumpName, LastSlash, bdaddr_str, strlen(bdaddr_str));
		}
	}


	// Save the actual file last.
	//
	save_binary_file(SafeName, LastSlash, body, body_len);





	// In BIP-Mode we need to answer with a Image-Handle
	//
	if( srv_Modes & OPDM_BIP ){
		char	ImageHandle[64];

		OBEX_ObjectSetRsp(object, OBEX_RSP_CONTINUE, OBEX_RSP_SUCCESS);

		// Generate next Image-Handle.
		//
		sprintf(ImageHandle,"%07d",srv_BIPNextImageHandle++);

		hv.bs = ImageHandle;
		OBEX_ObjectAddHeader(
			handle,
			object,
			0x30,				// Img-Handle
			hv,
			strlen(ImageHandle)+1,		// Length
			0
		);
	}else{
		OBEX_ObjectSetRsp(object, OBEX_RSP_CONTINUE, OBEX_RSP_SUCCESS);
	}
}













/* void Handle_OBEX_CMD_GET(obex_t *handle, obex_object_t *object)
**
**
**
** Handles a OBEX GET command.
**
**
** BIP Mode:
**	- Answer GetCapabilities() request.
**
** Non BIP:
**	- Answer NOT_FOUND to all.
**
*/
void Handle_OBEX_CMD_GET(obex_t *handle, obex_object_t *object)
{
	obex_headerdata_t	hv;
	uint8_t			hi;
	int			hlen;

	int			GotName=0;
	char			NameBuf[1024];

	int			GotType=0;
	char			TypeBuf[1024];



	printf("%s()\n", __FUNCTION__);

	while(OBEX_ObjectGetNextHeader(handle, object, &hi, &hv, &hlen)){
		switch(hi)	{
		case OBEX_HDR_NAME:
			printf("%s() Found name\n", __FUNCTION__);

			if( hlen/2 < sizeof(NameBuf) ){
				OBEX_UnicodeToChar(NameBuf, hv.bs, hlen);
				GotName = TRUE;
			}
			break;

		case OBEX_HDR_TYPE:
			printf("%s() HEADER_TYPE =  '%s' #%d  \n", __FUNCTION__, hv.bs, hlen);

			if( hlen < sizeof(TypeBuf) ){
				strcpy(TypeBuf,hv.bs);
				GotType = TRUE;
			}
			break;

		case 0xcb:
			printf("%s()CONNECTION_ID = %#x (%d length)\n", __FUNCTION__, hv.bq4,hlen);
			break;

		case 0x71:
			printf("%s()IMAGE_DESCRIPTION = ? (%d length)\n", __FUNCTION__, hlen);
			dump_hex((void *)hv.bs,hlen);
			break;


		default:
			printf("%s() Skipped header %02x\n", __FUNCTION__, hi);
		}
	}

	//
	//
	if( (srv_Modes & OPDM_BIP) && GotType && !strcmp(TypeBuf,"x-bt/img-capabilities") ){

		OBEX_ObjectSetRsp(object, OBEX_RSP_CONTINUE, OBEX_RSP_SUCCESS);

		hv.bs = bt_img_capabilities;
		OBEX_ObjectAddHeader(handle, object, OBEX_HDR_BODY, hv, strlen(bt_img_capabilities), 0);

		hv.bq4 = strlen(bt_img_capabilities);
		OBEX_ObjectAddHeader(handle, object, OBEX_HDR_LENGTH, hv, sizeof(uint32_t), 0);

		return;
	}



	// We won'tanswer any other get's.
	//
	if(GotName){
		printf("%s() Got a request for %s\n", __FUNCTION__, NameBuf);
	}else{
		printf("%s() Got a GET without a name-header!\n", __FUNCTION__);
	}

	OBEX_ObjectSetRsp(object, OBEX_RSP_NOT_FOUND, OBEX_RSP_NOT_FOUND);
	return;
}














/* void Handle_ServerRequest(obex_t *handle, obex_object_t *object, int event, int cmd)
 *
 *
 * Called when a request is about to come or has come.
 *
 */
void Handle_ServerRequest(obex_t *handle, obex_object_t *object, int event, int cmd)
{
	char	peeraddr_str[256];

	switch(cmd)	{
	case OBEX_CMD_SETPATH:
		printf("%s() Denied %02x request\n", __FUNCTION__, cmd);
		OBEX_ObjectSetRsp(object, OBEX_RSP_NOT_IMPLEMENTED, OBEX_RSP_NOT_IMPLEMENTED);
		//OBEX_ObjectSetRsp(object, OBEX_RSP_CONTINUE, OBEX_RSP_SUCCESS);
		break;

	case OBEX_CMD_PUT:
		Handle_OBEX_CMD_PUT(handle,object);
		break;

	case OBEX_CMD_GET:
		printf("%s() Denied %02x request\n", __FUNCTION__, cmd);
		OBEX_ObjectSetRsp(object, OBEX_RSP_NOT_IMPLEMENTED, OBEX_RSP_NOT_IMPLEMENTED);
		//Handle_OBEX_CMD_GET(handle,object);
		break;

	case OBEX_CMD_CONNECT:
		GetOBEXPeerName(handle,peeraddr_str);
		printf("peeraddr_str: %s\n", peeraddr_str);	
		printf("exists? : %d\n", db_dev_exists(peeraddr_str));	
		if (!db_dev_exists(peeraddr_str)) {
		    OBEX_ObjectSetRsp(object, OBEX_RSP_NOT_IMPLEMENTED, OBEX_RSP_NOT_IMPLEMENTED);
		    syslog(LOG_INFO, "OBEX connect attempt from %s",peeraddr_str);
		    break;
		}

		syslog(LOG_INFO, "OBEX connect from %s",peeraddr_str);

		if (!acquireLock(20)) { 
		    printf("timeout waiting for lock\n");
		    OBEX_ObjectSetRsp(object, OBEX_RSP_NOT_IMPLEMENTED, OBEX_RSP_NOT_IMPLEMENTED);
		    break;
		}
		
		printf("created lock\n");
		OBEX_ObjectSetRsp(object, OBEX_RSP_SUCCESS, OBEX_RSP_SUCCESS);
		break;

	case OBEX_CMD_DISCONNECT:
		GetOBEXPeerName(handle,peeraddr_str);
		syslog(LOG_INFO, "OBEX disconnect from %s",peeraddr_str);

		OBEX_ObjectSetRsp(object, OBEX_RSP_SUCCESS, OBEX_RSP_SUCCESS);
		unLock(); 
		printf("removed lock\n");
		break;

	default:
		printf("%s() Denied %02x request\n", __FUNCTION__, cmd);
		OBEX_ObjectSetRsp(object, OBEX_RSP_NOT_IMPLEMENTED, OBEX_RSP_NOT_IMPLEMENTED);
		break;
	}
	return;
}





char *event_name(int event)
{
	switch (event)	{
	case OBEX_EV_PROGRESS:		return "EV_PROGRESS";
	case OBEX_EV_REQDONE:		return "EV_REQDONE";
	case OBEX_EV_REQHINT:		return "EV_REQHINT";
	case OBEX_EV_REQ:		return "EV_REQ";
	case OBEX_EV_LINKERR:		return "EV_LINKERR";
	case OBEX_EV_PARSEERR:		return "EV_PARSEERR";
	case OBEX_EV_ABORT:		return "EV_ABORT";
	default:			return "EV_UNKNOWN";
	}
}

char *cmd_name(int cmd)
{
	switch(cmd)	{
	case OBEX_CMD_SETPATH:		return "CMD_SETPATH";
	case OBEX_CMD_PUT:		return "CMD_PUT";
	case OBEX_CMD_GET:		return "CMD_GET";
	case OBEX_CMD_CONNECT:		return "CMD_CONNECT";
	case OBEX_CMD_DISCONNECT:	return "CMD_DISCONNECT";
	case OBEX_CMD_ABORT:		return "CMD_ABORT";
	default:			return "CMD_??????";
	}
	return;
}




/*
 * Handle_OBEX_Event()
 *
 *    Called by the obex-layer when some event occurs.
 *
 */
void Handle_OBEX_Event(obex_t *handle, obex_object_t *object, int mode, int event, int obex_cmd, int obex_rsp)
{
	char	peeraddr_str[256];


	printf("\n		obex_event: %d  %d(%11s)  0x%x(%14s)   0x%x\n",
		mode,

		event,
		event_name(event),

		obex_cmd,
		cmd_name(obex_cmd),

		obex_rsp
	);



	//


	switch( event ){
	case OBEX_EV_PROGRESS:
		break;

	case OBEX_EV_REQDONE:
		break;

	case OBEX_EV_REQHINT:
		switch(obex_cmd) {
		case OBEX_CMD_PUT:
		case OBEX_CMD_CONNECT:
		case OBEX_CMD_DISCONNECT:
			OBEX_ObjectSetRsp(object, OBEX_RSP_CONTINUE, OBEX_RSP_SUCCESS);
			break;
		default:
			OBEX_ObjectSetRsp(object, OBEX_RSP_NOT_IMPLEMENTED, OBEX_RSP_NOT_IMPLEMENTED);
			break;
		}
		break;

	case OBEX_EV_REQ:
		Handle_ServerRequest(handle, object, event, obex_cmd);
		break;

	case OBEX_EV_ABORT:
		printf("abort!!\n");
		OBEX_ObjectSetRsp(object, OBEX_RSP_NOT_IMPLEMENTED, OBEX_RSP_NOT_IMPLEMENTED);
	case OBEX_EV_PARSEERR:
	case OBEX_EV_LINKERR:
		GetOBEXPeerName(handle,peeraddr_str);
		syslog(LOG_INFO, "lost link to %s",peeraddr_str);
		OBEX_TransportDisconnect(handle);
		unLock(); 
		printf("removed lock\n");
		break;

	default:
		printf("Unknown event %d !\n",event);
		break;
	}
}

